//
// main.cpp
//
// Console and command line code for disarm
// Taken mainly from Otfried Cheong's Arlo
// 21/01/00 Matt J. Gumbley
//
//

// for the command line editor
#include "..\..\boss\tuiedit\tuiedit.h"

#include <e32std.h>
#include <e32cons.h>
#include <f32file.h>
#include <e32svr.h>

#include "arlo.h"

typedef unsigned long int RawInstruction;

extern "C" char *disarm (RawInstruction rawinstruction);

#define VERSION "0.0a"
#define BYTES 1
#define HALFWORDS 2
#define WORDS 4

// public
LOCAL_D CConsoleBase* console; // write all your messages to this
LOCAL_D RFs fsSession;
LOCAL_D CLineEdit* commandLine;
LOCAL_D RArLo arlo;
LOCAL_D TUint32 *aPageMap;
LOCAL_D TInt aPageIdx;
TText8 *aOption;
GLDEF_D char hexdig[]="0123456789ABCDEF";
static char mybuf[30] = "here is my 30 %w%^%!ter stuff";
TInt LILO_enabled = 0;
TInt SNOOP_enabled = 0;
TInt MJG_enabled = 0;
TInt PHYS_enabled = 0;

LOCAL_C void help()
{
  console->Printf(_L("> d[x] [addr[,len]] : x is b|h|w, addr/len in hex\n"));
  console->Printf(_L("  dumps len bytes, halfwords or words from addr\n"));
  console->Printf(_L("> u [addr[,len]] : addr/len in hex\n"));
  console->Printf(_L("  disassembles len instructions from addr\n"));
  console->Printf(_L("> S : toggle SNOOP mode (reports lin->phy stuff)\n"));
  console->Printf(_L("> U : unload LILO.LDD forcibly (after crash)\n"));
  console->Printf(_L("> L : enable LILO.LDD, so you can do...\n"));
  console->Printf(_L("> P : toggle PHYS mode (all dumps done of phys. mem.)\n"));
  console->Printf(_L("> p <hex> : convert virtual to physical address\n"));
  console->Printf(_L("> v <hex> : convert physical to virtual address (may crash)\n"));
  console->Printf(_L("> e <hex> : examine word at physical address\n"));
  console->Printf(_L("> l <file> : load file\n"));
  console->Printf(_L("> o <options> : set boot options\n"));
  console->Printf(_L("> b : boot\n"));
  console->Printf(_L("> x : exit\n"));
  console->Printf(_L("> help1 for more help...\n"));
}

LOCAL_C void help1()
{
  console->Printf(_L("> M : toggle MJG mode (narrower dumps for Matt's duff Series 5)\n"));
  console->Printf(_L("> help for first page of help...\n"));
}
 
LOCAL_C TUint32 Linear2Physical(TUint32 aLinearAdr)
{
  TUint32 aPageDir = 0x40000014;
  // Get virtual address of page directory
  arlo.Examine(&aPageDir);
  if (SNOOP_enabled)
    console->Printf(_L("Page directory %08x\n"), aPageDir);
  TUint32 aLevelOne = aPageDir + ((aLinearAdr & 0xfff00000) >> 18);
  if (SNOOP_enabled)
    console->Printf(_L("Level One address %08x\n"), aLevelOne);
  arlo.Examine(&aLevelOne);
  if (SNOOP_enabled)
    console->Printf(_L("Level One descriptor %08x\n"), aLevelOne);
  // we now have the level one descriptor
  if ((aLevelOne & 0x3) == 0x2) {
    // a Section descriptor
    return ((aLevelOne & 0xfff00000) | (aLinearAdr & 0x000fffff));
  }
  if ((aLevelOne & 0x3) != 0x1) {
    // Fault or reserved
    if (SNOOP_enabled)
      console->Printf(_L("Level One fault or reserved\n"));
    return 0;
  }
  // Page descriptor.  Get Page table
  TUint32 aLevelTwo = (aLevelOne & 0xfffffc00);
  // convert Page table to linear
  aLevelTwo = RMmu::LinearAddress(aLevelTwo);
  aLevelTwo += ((aLinearAdr & 0x000ff000) >> 10);
  if (SNOOP_enabled)
    console->Printf(_L("Level Two address %08x\n"), aLevelTwo);
  arlo.Examine(&aLevelTwo);
  if (SNOOP_enabled)
    console->Printf(_L("Level Two descriptor %08x\n"), aLevelTwo);
  // level two descriptor
  if ((aLevelTwo & 0x3) == 0x2) {
    // a small (4K) page
    return ((aLevelTwo & 0xfffff000) | (aLinearAdr & 0x00000fff));
  }
  if ((aLevelTwo & 0x3) == 0x1) {
    // a large (64K) page
    return ((aLevelTwo & 0xffff0000) | (aLinearAdr & 0x0000ffff));
  }
  // fault or reserved
  if (SNOOP_enabled)
    console->Printf(_L("Level Two fault or reserved\n"));
  return 0;
}

LOCAL_C void loadFile(const TDesC &fileName)
{
  RFile aFile;
  console->Printf(_L("Opening file %S\n"), &fileName);
  TInt err = aFile.Open(fsSession, fileName, EFileRead);
  if (err) {
    console->Printf(_L("Error %d when opening file\n"), err);
    return;
  }
  console->Printf(_L("Succeeded\n"));
  TInt aSize;
  User::LeaveIfError(aFile.Size(aSize));
  console->Printf(_L("File size %x\n"), aSize);
  TUint8 *aBuf = new (ELeave) TUint8[aSize + 0x1000];
  // make it point to a page boundary
  TUint32 aLoadAdr = TUint32(aBuf);
  aLoadAdr += 0xfff;
  aLoadAdr &= 0xfffff000;
  aBuf = (TUint8 *) aLoadAdr;
  TPtr aFileBuf(aBuf, aSize);
  console->Printf(_L("Loading %x bytes to address %08x\n"), aSize, aLoadAdr);
  User::LeaveIfError(aFile.Read(aFileBuf));
  console->Printf(_L("Read %x bytes\n"), aFileBuf.Length());
  if (aFileBuf.Length() != aSize) 
    User::Leave(KErrNotFound);
  aFile.Close();

  TInt aBlockNo = (aSize + 0xfff) / 0x1000;    // No of 4k blocks needed
  console->Printf(_L("No pages = %d\n"), aBlockNo);

  if (aPageIdx + aBlockNo > 0x3fe) {
    console->Printf(_L("Does not fit in page table.\n"));
    return;
  }
  // add entries to page map 
  for (TInt i = 0; i < aBlockNo; i++) {
    aPageMap[aPageIdx++] = Linear2Physical(aLoadAdr + i * 0x1000);
  }
  aPageMap[aPageIdx++] = 0x0;
}


LOCAL_C unsigned char read_byte(TUint32 addr)
{
unsigned char *b;
TUint32 memword;
TInt mod;
TUint32 aVal;
  b = (unsigned char *)addr;

  if (PHYS_enabled) {
    memword = addr & ~0x03;
    mod = addr & 0x03;
    aVal = RMmu::LinearAddress(memword);
//console->Printf(_L("addr %08X mod %d av %08X\n"), memword, mod, aVal);
    arlo.Examine(&aVal);
    if (mod == 0) {
      return (unsigned char)(aVal & 0x000000ff);
    }
    else if (mod == 1) {
      return (unsigned char)((aVal & 0x0000ff00)>>8);
    }
    else if (mod == 2) {
      return (unsigned char)((aVal & 0x00ff0000)>>16);
    }
    else if (mod == 3) {
      return (unsigned char)((aVal & 0xff000000)>>24);
    }
    return 0xc9;
  }
  else {
    return b[0];
  }
}

LOCAL_C unsigned short int read_halfword(TUint32 addr)
{
unsigned short int *b;
TUint32 memword;
TInt mod;
TUint32 aVal;
  b = (unsigned short int *)addr;
  if (PHYS_enabled) {
    memword = addr & ~0x03;
    mod = addr & 0x03;
    aVal = RMmu::LinearAddress(memword);
    arlo.Examine(&aVal);
    if (mod == 0) {
      return (unsigned short int)(aVal & 0x0000ffff);
    }
    else if (mod == 2) {
      return (unsigned short int)((aVal & 0xffff0000)>>16);
    }
    return 0xc9c9;
  }
  else {
    return b[0];
  }
}

LOCAL_C TUint32 read_word(TUint32 exAdr)
{
TUint32 aVal;
unsigned long int *b;
  if (PHYS_enabled) {
    aVal = RMmu::LinearAddress(exAdr);
    arlo.Examine(&aVal);
    return aVal;
  }
  else {
    b = (unsigned long int *)exAdr;
    return (TUint32)*b;
  }
}

LOCAL_C void dump_bytes(TUint32 addr, TUint32 len)
{
TUint16 y,x;
TUint32 rows;
TBuf<67> LINE(_L(""));
TInt dumpwidth = MJG_enabled ? 8 : 16;
unsigned char *b, ch;

  rows = (len+(dumpwidth-1))/dumpwidth;
  b = (unsigned char *)addr;
  for (y=0; y<rows; y++) {
    console->Printf(_L("%08x "),b);
    LINE.Zero();
    LINE.Fill(' ',65);
    LINE.Append('\n');
    LINE.ZeroTerminate();
    for (x=0; x<dumpwidth; x++) {
      ch=read_byte(b+x);
      LINE[(3*x)]=hexdig[(ch & 0xf0)>>4];
      LINE[1+(3*x)]=hexdig[ch & 0x0f];
      TChar c=ch;
      LINE[(dumpwidth*3)+x]=c.IsPrint() ? ch : '.';
      len--;
      if (len == 0)
        break;
    }
    b += dumpwidth;
    console->Printf(_L("%S"), LINE);
    if (len == 0)
      break;
  }
}

LOCAL_C void dump_halfwords(TUint32 addr, TUint32 len)
{
TUint16 y,x;
TUint32 rows;
unsigned short int *b,hw;
TBuf<69> LINE(_L(""));

  rows = (len+7)/8;
  b = (unsigned short int *)addr;
  for (y=0; y<rows; y++) {
    console->Printf(_L("%08x "),b);
    LINE.Zero();
    LINE.Fill(' ',67);
    LINE.Append('\n');
    LINE.ZeroTerminate();

    for (x=0; x<8; x++) {
      hw=read_halfword(b+x);
      LINE[0+(5*x)]=hexdig[(hw&0xf000)>>12];
      LINE[1+(5*x)]=hexdig[(hw&0x0f00)>>8];
      LINE[2+(5*x)]=hexdig[(hw&0x00f0)>>4];
      LINE[3+(5*x)]=hexdig[(hw&0x000f)];
      len--;
      if (len == 0)
        break;
    }
    b += 0x08;
    console->Printf(_L("%S"), LINE);
    if (len == 0)
      break;
  }
}

LOCAL_C void dump_words(TUint32 addr, TUint32 len)
{
TUint16 y,x;
TUint32 rows;
unsigned long int *b, w;
TBuf<69> LINE(_L(""));

  rows = (len+3)/4;
  addr &= ~0x03;
  b = (unsigned long int *)addr;
  for (y=0; y<rows; y++) {
    console->Printf(_L("%08x "),b);
    LINE.Zero();
    LINE.Fill(' ',67);
    LINE.Append('\n');
    LINE.ZeroTerminate();

    for (x=0; x<4; x++) {
      w = read_word(b+x);
      LINE[0+(9*x)]=hexdig[(w&0xf0000000)>>28];
      LINE[1+(9*x)]=hexdig[(w&0x0f000000)>>24];
      LINE[2+(9*x)]=hexdig[(w&0x00f00000)>>20];
      LINE[3+(9*x)]=hexdig[(w&0x000f0000)>>16];
      LINE[4+(9*x)]=hexdig[(w&0x0000f000)>>12];
      LINE[5+(9*x)]=hexdig[(w&0x00000f00)>>8];
      LINE[6+(9*x)]=hexdig[(w&0x000000f0)>>4];
      LINE[7+(9*x)]=hexdig[(w&0x0000000f)];
      len--;
      if (len == 0)
        break;
    }
    b += 0x04;
    console->Printf(_L("%S"), LINE);
    if (len == 0)
      break;
  }
}


LOCAL_C void dump(TUint32 type, TUint32 addr, TUint32 len)
{
  if (type == BYTES)
    dump_bytes(addr, len);
  if (type == HALFWORDS)
    dump_halfwords(addr, len);
  if (type == WORDS)
    dump_words(addr, len);
}

LOCAL_C void disassemble(TUint32 addr, TUint32 len)
{
unsigned long int *b;

  b = (unsigned long int *)addr;
  RawInstruction rawinst;
  while (len > 0) {
    rawinst = read_word(b);
    console->Printf(_L("%08x %s\n"), b, disarm(rawinst));
    b++;
    len--;
  }
}

LOCAL_C void enable_LILO()
{
  // load the driver
  console->Printf(_L("Loading logical device\n"));
  // searches system path for *.ldd (logical device driver)
  TInt err = User::LoadLogicalDevice(_L("LILO"));
  if (err) {
    console->Printf(_L("Error %d while loading logical device\n"), err);
    return;
  }
  // open a channel
  console->Printf(_L("Opening the channel\n"));
  err = arlo.Open(arlo.VersionRequired());
  if (err) {
    console->Printf(_L("Error %d while opening channel\n"), err);
    return;
  }

  // prepare page map
  aPageMap = new (ELeave) TUint32[0x800]; // 2k entries
  // adjust to 4k page
  TUint32 aAdr = TUint32(aPageMap);
  aAdr += 0xfff;
  aAdr &= 0xfffff000;
  aPageMap = (TUint32 *) aAdr;

  // clear page map
  for (TInt i = 0; i < 0x400; i++)
    aPageMap[i] = 0x0;
  // prepare option string buffer
  aOption = (TText8 *) new (ELeave) TUint32[64]; // word align 
  aOption[0] = '\0';
  // first entry in page map:  phys address of option string
  aPageMap[0] = Linear2Physical(TUint32(aOption));
  console->Printf(_L("Option buffer at address %08x\n"), aPageMap[0]);
  // second entry: phys address of memory description
  aAdr = 0x40000030;
  arlo.Examine(&aAdr);
  aPageMap[1] = Linear2Physical(aAdr);
  console->Printf(_L("Memory description at address %08x\n"), aPageMap[1]);
  aPageIdx = 2;

  LILO_enabled = 1;
}

LOCAL_C void murder_LILO()
{
  console->Printf(_L("Freeing the device\n"));
  User::LeaveIfError(User::FreeLogicalDevice(_L("ArLo")));
  User::FreeLogicalDevice(_L("ArLo"));
}

LOCAL_C void doExampleL()
{
 
  TBuf<80> cline;
  console->Printf(_L("disarm v%s (C) 2000 Matt J. Gumbley\n"), VERSION);
  console->Printf(_L("Portions (C) Kevin F. Quinn & Otfried Cheong.\n"));
  console->Printf(_L("Type 'help' for a list of commands. 'x' to exit.\n"));
  console->Printf(_L("\n"));
  
 
  TBool isFinished = EFalse;
  TUint32 aVal;
  TUint32 dumptype = BYTES;
  TUint32 dumplen = 16;
  TUint32 dumpaddr = 0;
  TUint32 exAdr = 0;
  console->Printf(_L("mybuf is at %08x\n"), &mybuf[0]);
  console->Printf(_L("dump_words is at %08x\n"), &dump_words);
  do {
    commandLine->Edit(_L("> "), &cline);
    
    TLex lex(cline);
    TPtrC command = lex.NextToken();
    lex.SkipSpace();
    if (command == _L("help")) {
      help();
      continue;
    } 
    if (command == _L("help1")) {
      help1();
      continue;
    } 

    if (command == _L("db"))
      dumptype = BYTES;
    if (command == _L("dh"))
      dumptype = HALFWORDS;
    if (command == _L("dw"))
      dumptype = WORDS;

    if (command == _L("db") || command == _L("dh") || command == _L("dw") ||
        command == _L("d")) {
      if (lex.Val(aVal, EHex) == KErrNone) {
        dumpaddr = aVal;
        lex.SkipSpace();
        if (lex.Val(aVal, EHex) == KErrNone) {
          dumplen = aVal;
        }
      }
      else {
        dumpaddr += (dumplen * dumptype);
      }
      dump(dumptype, dumpaddr, dumplen);
    }

    else if (command == _L("u")) {
      if (lex.Val(aVal, EHex) == KErrNone) {
        dumpaddr = aVal;
        lex.SkipSpace();
        if (lex.Val(aVal, EHex) == KErrNone) {
          dumplen = aVal;
        }
      }
      else {
        dumpaddr += dumplen;
      }
      disassemble(dumpaddr, dumplen);

    } else if (command == _L("p") || command == _L("v") ) {
      if (!LILO_enabled) {
        console->Printf(_L("LILO not initialised (use 'L')\n"));
        continue;
      }
      if (lex.Val(aVal, EHex) != KErrNone) {
        console->Printf(_L("Illegal number\n"));
        continue;
      }
      if (command == _L("p")) {
        TUint32 aPhysAdr = Linear2Physical(aVal);
        console->Printf(_L("%08X -> Physical %08X\n"), aVal, aPhysAdr);
      } else if (command == _L("v")) {
        TUint32 aLinAdr = RMmu::LinearAddress(aVal);
        console->Printf(_L("%08X -> Virtual  %08X\n"), aVal, aLinAdr);
      }

    } else if (command == _L("e")) {
      if (!LILO_enabled) {
        console->Printf(_L("LILO not initialised (use 'L')\n"));
        continue;
      }
      if (lex.Val(aVal, EHex) == KErrNone) {
        exAdr = aVal;
      } else {
      	exAdr += 4;
      }
      aVal = RMmu::LinearAddress(exAdr);
      arlo.Examine(&aVal);
      console->Printf(_L("%08X: %08X\n"), exAdr, aVal);

    } else if (command == _L("f")) {
      // might be the same as Otfried's Linear2Physical?
      if (lex.Val(aVal, EHex) == KErrNone) {
        exAdr = aVal;
      } else {
      	exAdr += 4;
      }
      aVal = RMmu::PhysicalAddress(exAdr);
      console->Printf(_L("%08X -> Physical(EPOC) %08X\n"), exAdr, aVal);


    } else if (command == _L("b")) {
      if (!LILO_enabled) {
        console->Printf(_L("LILO not initialised (use 'L')\n"));
        continue;
      }
      if (aPageIdx == 2) {
	console->Printf(_L("no file loaded\n"));
      } else {
	// boot
	TUint32 aPhysMapAdr = Linear2Physical(TUint32(aPageMap));
	// Boot: pass phys adr of map, and phys adr of code
	arlo.Boot(aPhysMapAdr, aPageMap[2]);
	// never come back here
      }

    } else if (command == _L("o")) {
      // copy option string to option buffer
      lex.SkipSpace();
      TPtrC anOpt = lex.Remainder();
      if (anOpt.Length() > 255) {
	console->Printf(_L("String is too long\n"));
      } else {
	TText8 *p = aOption;
	for (TInt i = 0; i < anOpt.Length(); i++)
	  *p++ = anOpt[i];
	*p++ = '\0';
      }

    } else if (command == _L("l")) {
      if (!LILO_enabled) {
        console->Printf(_L("LILO not initialised (use 'L')\n"));
        continue;
      }
      // load a file
      TPtrC bootName = lex.NextToken();
      loadFile(bootName);

    } else if (command == _L("M")) {
      MJG_enabled = MJG_enabled ^ 0x01;
      console->Printf(_L("MJG mode %s\n"), MJG_enabled ? "enabled" : "disabled");

    } else if (command == _L("S")) {
      SNOOP_enabled = SNOOP_enabled ^ 0x01;
      console->Printf(_L("SNOOP mode %s\n"), SNOOP_enabled ? "enabled" : "disabled");

    } else if (command == _L("P")) {
      if (!LILO_enabled) {
        console->Printf(_L("LILO not initialised (use 'L')\n"));
        continue;
      }
      PHYS_enabled = PHYS_enabled ^ 0x01;
      console->Printf(_L("PHYS mode %s\n"), PHYS_enabled ? "enabled" : "disabled");

    } else if (command == _L("L")) {
      enable_LILO();

    } else if (command == _L("U")) {
      murder_LILO();

    } else if (command == _L("x")) {
      isFinished = ETrue;

    } else {
      console->Printf(_L("Unknown command\n"));
    }
  } while (!isFinished);

  // ok, that's all
  if (LILO_enabled) {
     console->Printf(_L("Closing the channel\n"));
     arlo.Close();
     console->Printf(_L("Freeing the device\n"));
     User::LeaveIfError(User::FreeLogicalDevice(_L("ArLo")));
     // refers to device driver by loaded name - not filename
  }
}

LOCAL_C void callExampleL() // initialize and call under cleanup stack
{
  console=Console::NewL(_L("DISARM"),
			TSize(KDefaultConsWidth,KDefaultConsHeight));
  CleanupStack::PushL(console);
  commandLine=CLineEdit::NewL(console,10);
  TInt fsret = fsSession.Connect(); // start a file session
  if (fsret != KErrNone) {
    console->Printf(_L("file session connect failed: return code=%d"),fsret);
    User::Leave(fsret);
  }
  TRAPD(error,doExampleL()); // perform example function
  if (error) {
    if (LILO_enabled) 
      User::FreeLogicalDevice(_L("ArLo"));
    console->Printf(_L("failed: leave code=%d\n"), error);
    console->Printf(_L("Press a key\n"));
    console->Getch();
  }
  fsSession.Close(); //close the file session
  CleanupStack::PopAndDestroy(); // close console
}


GLDEF_C TInt E32Main() // main function called by E32
{
  __UHEAP_MARK;
  CTrapCleanup* cleanup=CTrapCleanup::New();       
  TRAPD(error,callExampleL());                     
  __ASSERT_ALWAYS(!error,User::Panic(_L("DISARM"),error));
  delete cleanup; // destroy clean-up stack
  __UHEAP_MARKEND;
  return 0; // and return
}

